use std::path::PathBuf;

use super::{Package, arch::Architecture};

/// Package Repository
pub struct Repository {
    pub name: String,
}

impl Repository {
    /// Get a list of all package repositories
    pub fn list() -> Vec<String> {
        let mut repos = vec![];

        for entry in std::fs::read_dir("./data").unwrap() {
            let path = entry.unwrap().path();
            let file_name = path.file_name().unwrap().to_str().unwrap().to_string();
            repos.push(file_name);
        }

        repos
    }

    /// Create a new package repository with architectures from `arch`
    pub fn create(name: &str, arch: Vec<Architecture>) -> Repository {
        let path = PathBuf::from("./data").join(name);
        std::fs::create_dir_all(&path).unwrap();

        for arch in arch {
            let np = path.join(arch.to_string());
            std::fs::create_dir_all(np).unwrap();
        }

        Repository::new(name).unwrap()
    }
}

impl Repository {
    /// Get a repository if it exists
    pub fn new(name: &str) -> Option<Self> {
        if PathBuf::from("./data").join(name).exists() {
            Some(Repository {
                name: name.to_string(),
            })
        } else {
            None
        }
    }

    /// Get a list of architectures for this repository
    pub fn arch(&self) -> Vec<Architecture> {
        let dir_path = PathBuf::from("./data").join(&self.name);
        let mut arch = vec![];

        if let Ok(entries) = std::fs::read_dir(dir_path) {
            for entry in entries.filter_map(Result::ok) {
                let file_name = entry.file_name().into_string().unwrap_or_default();
                if let Some(repo_arch) = Architecture::parse(&file_name) {
                    arch.push(repo_arch);
                }
            }
        }

        arch
    }

    /// Get the base path for the repository with `arch`.
    pub fn base_path(&self, arch: Architecture) -> PathBuf {
        PathBuf::from("./data")
            .join(&self.name)
            .join(&arch.to_string())
    }

    /// Get the `.db.tar.gz` content for the repository of `arch`
    pub fn db_content(&self, arch: Architecture) -> Option<Vec<u8>> {
        std::fs::read(
            self.base_path(arch)
                .join(format!("{}.db.tar.gz", self.name)),
        )
        .ok()
    }

    /// Get the `.db.tar.gz.sig` content for the repository of `arch`
    pub fn sig_content(&self, arch: Architecture) -> Option<Vec<u8>> {
        std::fs::read(
            self.base_path(arch)
                .join(format!("{}.db.tar.gz.sig", self.name)),
        )
        .ok()
    }

    pub fn get_pkg(&self, pkg_name: &str) -> Option<Package> {
        // Normalize name
        let pkg_name = if pkg_name.ends_with(".sig") {
            pkg_name.trim_end_matches(".sig").to_string()
        } else {
            pkg_name.to_string()
        };

        // Find package
        let (name, version, _, arch) = Package::extract_pkg_name(&pkg_name).unwrap();
        let pkg = Package::new(&self.name, arch, &name, &version);

        // Return if exists
        if pkg.exists() {
            return Some(pkg);
        } else {
            None
        }
    }
}
